@@ -1,12 +1,15 @@
 import random
 import transaction
 import types
+import codecs
+import os
 import sqlalchemy as sa
 import velruse
 import json as _json
 import StringIO
 import qrcode as qrcode_module
 import docutils.examples
+import markupsafe
 from datetime import datetime
 
 from mako.template import Template as t
@@ -39,6 +42,9 @@ import widgets
 
 from moksha.wsgi.widgets.api import get_moksha_socket, LiveWidget
 
+from sqlalchemy.orm import joinedload
+from sqlalchemy.sql.expression import func
+
 # Optional.  Emit messages to the fedmsg bus.
 fedmsg = None
 try:
@@ -46,6 +52,21 @@ try:
 except ImportError:
     pass
 
+def _get_user(request, id_or_nickname):
+    '''Attempt to get a user by their id or nickname, returning None if
+       we fail.'''
+    user = request.db.get_person(nickname=id_or_nickname)
+
+    if user:
+        return user
+    else:
+        try:
+            # We cast user_id to an integer so that Postgres doesn't
+            # get upset about comparing what is potentially a string
+            # to an integer column.
+            return request.db.get_person(id=int(id_or_nickname))
+        except ValueError:
+            return None
 
 @view_config(route_name='admin', renderer='admin.mak', permission='admin')
 def admin(request):
@@ -292,41 +313,51 @@ def leaderboard(request):
     else:
         awarded_assertions = None
 
-    # Get top persons.
-    persons_assertions = request.db.get_all_assertions().join(m.Person)
-    from collections import defaultdict
-    top_persons = defaultdict(int) # person: assertion count
-    for item in persons_assertions:
-        top_persons[item.person] += 1
+    leaderboard = request.db.session.query(m.Person, func.count(
+        m.Person.assertions)).join(m.Assertion).order_by(
+            'count_1 desc').filter(m.Person.opt_out == False).group_by(
+                m.Person).all()
 
-    # top_persons and top_persons_sorted contain all persons, ordered
-    top_persons_sorted = sorted(sorted(top_persons,
-                                key=lambda person: person.id),
-                                key=top_persons.get,
-                                reverse=True)
+    # Hackishly, but relatively cheaply get the rank of all users.
+    # This is:
+    # { <person object>:
+    #   {
+    #     'badges': <number of badges they have>,
+    #     'rank': <their global rank>
+    #   }
+    # }
+    user_to_rank = dict(
+        [
+            [
+                data[0],
+                {
+                    'badges': data[1],
+                    'rank': idx + 1
+                }
+            ] for idx, data in enumerate(leaderboard)
+        ]
+    )
 
     # Get total user count.
-    user_count = len(top_persons)
+    user_count = len(leaderboard)
 
     if authenticated_userid(request):
+        user = request.db.get_person(
+            person_email=authenticated_userid(request))
+
         # Get rank.
+        idx = [i[0] for i in leaderboard].index(user)
+        rank = idx + 1
+        # Handle the case of leaderboard[-2:2] which will be [] always.
+        if idx < 2:
+            idx = 2
+        competitors = [c[0] for c in leaderboard[(idx - 2):(idx + 3)]]
+
         try:
-            rank = top_persons_sorted.index(request.db.get_person(
-                                person_email=authenticated_userid(
-                                             request))) + 1
-        except ValueError:
-            rank = 0
-        # Get percentile.
-        try:
-            percentile = (float(rank) / float(user_count)) * 100
+            percentile = (float(user_to_rank[user]['rank']) / float(user_count)) * 100
         except ZeroDivisionError:
             percentile = 0
 
-        # Get a list of nearby competetors (5 users above the current
-        # user and 5 users ranked below).
-        competitors = top_persons_sorted[max(rank - 3, 0):\
-                                     min(rank + 2, len(top_persons_sorted))]
-
     else:
         rank = None
         percentile = None
@@ -335,44 +366,99 @@ def leaderboard(request):
     return dict(
             auth_principals=effective_principals(request),
             awarded_assertions=awarded_assertions,
-            top_persons=top_persons,
-            top_persons_sorted=top_persons_sorted,
+            top_persons_sorted=[x[0] for x in leaderboard],
             rank=rank,
             user_count=user_count,
             percentile=percentile,
             competitors=competitors,
+            user_to_rank=user_to_rank,
             )
 
+@view_config(route_name='leaderboard_json', renderer='json')
+@view_config(route_name='rank_json', renderer='json')
+def leaderboard_json(request):
+    """ Render a top-users JSON dump. """
+
+    user_id = request.matchdict.get('id')
+    user = None
+    if user_id:
+        user = _get_user(request, user_id)
+
+    leaderboard = request.db.session.query(m.Person, func.count(
+        m.Person.assertions)).join(m.Assertion).order_by(
+            'count_1 desc').filter(m.Person.opt_out == False).group_by(
+                m.Person).all()
+
+    # Hackishly, but relatively cheaply get the rank of all users.
+    # This is:
+    # { <person object>:
+    #   {
+    #     'badges': <number of badges they have>,
+    #     'rank': <their global rank>
+    #   }
+    # }
+    user_to_rank = dict(
+        [
+            [
+                data[0],
+                {
+                    'badges': data[1],
+                    'rank': idx + 1
+                }
+            ] for idx, data in enumerate(leaderboard)
+        ]
+    )
+
+    if user:
+        idx = [i[0] for i in leaderboard].index(user)
+        # Handle the case of leaderboard[-2:2] which will be [] always.
+        if idx < 2:
+            idx = 2
+        leaderboard = leaderboard[(idx - 2):(idx + 3)]
+    else:
+        leaderboard = leaderboard[:25]
+
+    ret = [
+        dict(user_to_rank[p[0]].items() + {'nickname': p[0].nickname}.items())
+        for p in leaderboard]
+
+    return { 'leaderboard': ret }
+
+
+@view_config(route_name='about', renderer='about.mak')
+def about(request):
+    return dict(content=load_docs(request, 'about'))
+
 
 @view_config(route_name='explore', renderer='explore.mak')
 def explore(request):
 
     # Check if a search has been done, and if so, show
     # search results.
+    search_query = None
     search_results = dict() # name: link
     if request.POST:
         if request.POST.get('badge-search'):
             # badge-query is a required field on the template form.
-            badge_query = request.POST.get('badge-query')
+            search_query = request.POST.get('badge-query')
             matching_results = request.db.get_all_badges().filter(
-                    sa.func.lower((m.Badge.name.like('%' + badge_query
-                                    + '%'))) |
-                    sa.func.lower((m.Badge.description.like('%' +
-                                    badge_query +
-                                    '%'))) |
-                    sa.func.lower((m.Badge.tags.like('%' +
-                                    badge_query
-                                    + '%')))).all()
+                    sa.func.lower(m.Badge.name).like('%' + search_query
+                                    + '%') |
+                    sa.func.lower(m.Badge.description).like('%' +
+                                    search_query +
+                                    '%') |
+                    sa.func.lower(m.Badge.tags).like('%' +
+                                    search_query
+                                    + '%')).all()
             for r in matching_results:
-                search_results[r.name] = request.route_url('badge',
-                        id=r.name.lower().replace(' ', '-'))
+                search_results[r.name] = request.route_url('badge', id=r.id)
         elif request.POST.get('person-search'):
             # person-query is a required field on the template form.
-            person_query = request.POST.get('person-query')
+            search_query = request.POST.get('person-query')
             matching_results = request.db.get_all_persons().filter(
-                    ((m.Person.nickname.like('%' + person_query
+                    ((m.Person.nickname.like('%' + search_query
                             + '%')) |
-                    (m.Person.bio.like('%' + person_query
+                    (m.Person.bio.like('%' + search_query
                             + '%'))) &
                     (m.Person.opt_out == False)).all()
             for r in matching_results:
@@ -380,13 +466,13 @@ def explore(request):
                         'user', id=r.nickname)
         elif request.POST.get('tag-search'):
             # tag-query is a required field on the template form.
-            tag_query = request.POST.get('tag-query')
+            search_query = request.POST.get('tag-query')
             if request.POST.get('tag-match-all'):
                 return HTTPFound(location=request.route_url(
-                                 'tags', tags=tag_query, match='all'))
+                                 'tags', tags=search_query, match='all'))
             else:
                 return HTTPFound(location=request.route_url(
-                                 'tags', tags=tag_query, match='any'))
+                                 'tags', tags=search_query, match='any'))
 
 
     # Get awarded assertions.
@@ -416,9 +502,33 @@ def explore(request):
             random_badges=random_badges,
             random_persons=random_persons,
             search_results=search_results,
+            search_query=search_query,
             )
 
 
+@view_config(route_name='explore_badges', renderer='explore_badges.mak')
+def explore_badges(request):
+
+    all_badges = request.db.get_all_badges().all()
+
+    newest_badges = sorted(request.db.get_all_badges().all(),
+                           key=lambda badge: badge.created_on,
+                           reverse=True)[:20]
+
+    # Get awarded assertions.
+    if authenticated_userid(request):
+        awarded_assertions = request.db.get_assertions_by_email(
+                                authenticated_userid(request))
+    else:
+        awarded_assertions = None
+
+    return dict(
+            all_badges=all_badges,
+            newest_badges=newest_badges,
+            auth_principals=effective_principals(request),
+            awarded_assertions=awarded_assertions,
+            )
+
 @view_config(route_name='badge', renderer='badge.mak')
 def badge(request):
     """Render badge info page."""
@@ -457,12 +567,17 @@ def badge(request):
 
         percent_earned = float(times_awarded) / \
                          float(len(request.db.get_all_persons().all()))
+
+        # This is a list of assertions for this badge.
+        badge_assertions = request.db.get_all_assertions().filter(
+                m.Assertion.badge_id == badge.id).all()
     except sa.orm.exc.NoResultFound: # This badge has never been awarded.
         times_awarded = 0
         last_awarded = None
         last_awarded_person = None
         first_awarded = None
         first_awarded_person = None
+        badge_assertions = None
         percent_earned = 0
     # Percent of people who have earned this badge
 
@@ -481,6 +596,7 @@ def badge(request):
             last_awarded_person=last_awarded_person,
             first_awarded=first_awarded,
             first_awarded_person=first_awarded_person,
+            badge_assertions=badge_assertions,
             percent_earned=percent_earned,
             )
 
@@ -573,23 +689,9 @@ def user(request):
     else:
         awarded_assertions = None
 
-    # So, here they can use their 'id' or their 'nickname'.
-    # We'll try nickname first since we want to encourage that (or whatever)
-    # and fall back to id if that fails.  If both fail, raise a 404.
     user_id = request.matchdict.get('id')
+    user = _get_user(request, user_id)
 
-    user = request.db.get_person(nickname=user_id)
-
-    if not user:
-        try:
-            # We cast user_id to an integer so that Postgres doesn't
-            # get upset about comparing what is potentially a string
-            # to an integer column.
-            user = request.db.get_person(id=int(user_id))
-        except TypeError:
-            raise HTTPNotFound("No such user %r" % user_id)
-
-    # If we still haven't found anything, just give up.
     if not user:
         raise HTTPNotFound("No such user %r" % user_id)
 
@@ -636,6 +738,27 @@ def user(request):
     invitations = [i for i in request.db.get_invitations(user.id)\
                    if i.expires_on > datetime.now()]
 
+    # Get rank. (same code found in leaderboard view function)
+    persons_assertions = request.db.get_all_assertions().join(m.Person).filter(
+        m.Person.opt_out == False)
+    from collections import defaultdict
+    top_persons = defaultdict(int) # person: assertion count
+    for item in persons_assertions:
+        top_persons[item.person] += 1
+    top_persons_sorted = sorted(sorted(top_persons,
+                                key=lambda person: person.id),
+                                key=top_persons.get,
+                                reverse=True)
+    user_count = len(top_persons)
+    try:
+        rank = top_persons_sorted.index(user) + 1
+    except ValueError:
+        rank = 0
+    try:
+        percentile = (float(rank) / float(user_count)) * 100
+    except ZeroDivisionError:
+        percentile = 0
+
     return dict(
             user=user,
             user_badges=user_badges,
@@ -644,39 +767,12 @@ def user(request):
             auth_principals=effective_principals(request),
             awarded_assertions=awarded_assertions,
             allow_changenick=allow_changenick,
+            rank=rank,
+            percentile=percentile,
+            user_count=user_count,
             )
 
-
-@view_config(route_name='user_json', renderer='json')
-def user_json(request):
-    """Render user info JSON dump."""
-
-    # So, here they can use their 'id' or their 'nickname'.
-    # We'll try nickname first since we want to encourage that (or whatever)
-    # and fall back to id if that fails.  If both fail, raise a 404.
-    user_id = request.matchdict.get('id')
-
-    user = request.db.get_person(nickname=user_id)
-
-    if not user:
-        try:
-            # We cast user_id to an integer so that Postgres doesn't
-            # get upset about comparing what is potentially a string
-            # to an integer column.
-            user = request.db.get_person(id=int(user_id))
-        except TypeError:
-            request.response.status = '404 Not Found'
-            return {"error": "No such user exists."}
-
-    # If we still haven't found anything, just give up.
-    if not user:
-        request.response.status = '404 Not Found'
-        return {"error": "No such user exists."}
-
-    if user.opt_out == True and user.email != authenticated_userid(request):
-        request.response.status = '404 Not Found'
-        return {"error": "User has opted out."}
-
+def _user_json_generator(request, user):
     awarded_assertions = request.db.get_assertions_by_email(user.email)
 
     # Get user badges.
@@ -710,6 +806,25 @@ def user_json(request):
         'assertions': assertions
     }
 
+@view_config(route_name='user_json', renderer='json')
+def user_json(request):
+    """Render user info JSON dump."""
+
+    # So, here they can use their 'id' or their 'nickname'.
+    # We'll try nickname first since we want to encourage that (or whatever)
+    # and fall back to id if that fails.  If both fail, raise a 404.
+    user = _get_user(request, request.matchdict.get('id'))
+
+    if not user:
+        request.response.status = '404 Not Found'
+        return {"error": "No such user exists."}
+
+    if user.opt_out == True and user.email != authenticated_userid(request):
+        request.response.status = '404 Not Found'
+        return {"error": "User has opted out."}
+
+    return _user_json_generator(request, user)
+
 @view_config(route_name='builder', renderer='builder.mak')
 def builder(request):
     if authenticated_userid(request):
@@ -895,3 +1010,78 @@ def make_websocket_handler(settings):
         template = ""
 
     return WebsocketHandler
+
+
+def modify_rst(rst):
+    """ Downgrade some of our rst directives if docutils is too old. """
+
+    try:
+        # The rst features we need were introduced in this version
+        minimum = [0, 9]
+        version = map(int, docutils.__version__.split('.'))
+
+        # If we're at or later than that version, no need to downgrade
+        if version >= minimum:
+            return rst
+    except Exception:
+        # If there was some error parsing or comparing versions, run the
+        # substitutions just to be safe.
+        pass
+
+    # Otherwise, make code-blocks into just literal blocks.
+    substitutions = {
+        '.. code-block:: javascript': '::',
+    }
+    for old, new in substitutions.items():
+        rst = rst.replace(old, new)
+
+    return rst
+
+
+def modify_html(html):
+    """ Perform style substitutions where docutils doesn't do what we want.
+    """
+
+    substitutions = {
+        '<tt class="docutils literal">': '<code>',
+        '</tt>': '</code>',
+    }
+    for old, new in substitutions.items():
+        html = html.replace(old, new)
+
+    return html
+
+
+def _load_docs(directory, endpoint):
+    """ Utility to load an RST file and turn it into fancy HTML. """
+
+    fname = os.path.join(directory, endpoint + '.rst')
+    with codecs.open(fname, 'r', 'utf-8') as f:
+        rst = f.read()
+
+    rst = modify_rst(rst)
+
+    api_docs = docutils.examples.html_body(rst)
+
+    api_docs = modify_html(api_docs)
+
+    api_docs = markupsafe.Markup(api_docs)
+    return api_docs
+
+
+htmldocs = {}
+def load_docs(request, key):
+    possible_keys = ['about', 'footer']
+
+    # Load from disk only once on first request.
+    if not htmldocs:
+        here = os.path.dirname(os.path.abspath(__file__))
+        dflt = os.path.join(here, 'sitedocs')
+        directory = request.registry.settings.get('tahrir.sitedocs_dir', dflt)
+        for k in possible_keys:
+            htmldocs[k] = _load_docs(directory, k)
+
+    if key not in htmldocs:
+        raise KeyError("%r is not permitted." % key)
+
+    return htmldocs[key]
